package edu.umass.cs.flux;

import java_cup.runtime.Symbol;

%%
%cup
%eofval{
	if (yy_lexical_state == COMMENT)
		throw(new Error("Error: Unmatched start of comment at line " + comment_line + "."));
  return new Symbol(sym.EOF);
%eofval}
%line
%state COMMENT

COMMENT_TEXT=([^*]|"*"[^/])*

%{
	private int comment_line = 0;

  public int getLine() {
    return yyline + 1;
  }
%}
%%
<YYINITIAL>	"=>"	{ return new Symbol(sym.ARROW); }
<YYINITIAL>	"("		{ return new Symbol(sym.LEFT_PAREN); }
<YYINITIAL>	")"		{ return new Symbol(sym.RIGHT_PAREN); }
<YYINITIAL>	"["		{ return new Symbol(sym.LEFT_SQ_BRACE); }
<YYINITIAL>	"]"		{ return new Symbol(sym.RIGHT_SQ_BRACE); }
<YYINITIAL>	"{"		{ return new Symbol(sym.LEFT_CR_BRACE); }
<YYINITIAL>	"}"		{ return new Symbol(sym.RIGHT_CR_BRACE); }
<YYINITIAL>	"->"	{ return new Symbol(sym.PIPE); }
<YYINITIAL>	":"		{ return new Symbol(sym.COLON); }
<YYINITIAL>	";"		{ return new Symbol(sym.SEMI); }
<YYINITIAL>	","		{ return new Symbol(sym.COMMA); }
<YYINITIAL>	"="		{ return new Symbol(sym.EQUALS); }
<YYINITIAL>	"*"		{ return new Symbol(sym.STAR); }
<YYINITIAL>	"_"		{ return new Symbol(sym.UNDERSCORE); }
<YYINITIAL>	"+"		{ return new Symbol(sym.PLUS); }
<YYINITIAL>	"?"		{ return new Symbol(sym.QUESTION); }
<YYINITIAL>	"!"		{ return new Symbol(sym.EXCLAMATION); }

<YYINITIAL> "/*"			{ yybegin(COMMENT); comment_line = yyline + 1; }
<COMMENT>	{COMMENT_TEXT}	{ }
<COMMENT>	"*/"			{ yybegin(YYINITIAL); }
<YYINITIAL>	"//".*			{ }

<YYINITIAL>	"source"	{ return new Symbol(sym.SOURCEDEF); }
<YYINITIAL>	"handle"	{ return new Symbol(sym.HANDLE); }
<YYINITIAL>	"error"		{ return new Symbol(sym.ERROR); }
<YYINITIAL>	"atomic"	{ return new Symbol(sym.ATOMIC); }
<YYINITIAL>	"program"	{ return new Symbol(sym.PROGRAM); }
<YYINITIAL>	"session"	{ return new Symbol(sym.SESSION); }

<YYINITIAL>	("typedef"|"predicate")	{ return new Symbol(sym.TYPEDEF); }

<YYINITIAL> [0-9]+
	{ return new Symbol(sym.NUMBER, new Integer(yytext())); }
<YYINITIAL> [a-zA-Z_*][a-zA-Z0-9_*]* 
	{ return new Symbol(sym.IDENTIFIER, new String(yytext()));}
<YYINITIAL, COMMENT> [ \t\r\n\f]
	{ /* ignore white space. */ }
<YYINITIAL, COMMENT> .
	{ System.err.println("Illegal character: "+yytext()); }
